// Scatter Gun patch. Originally intended for use on the
// Unknown Enemy project, but later dropped and more or
// less forgotten...
// 
// As it stands, the patch simulates the WC4 Scatter gun by
// splitting the Cloudburst bolt in 5 different bolts, with
// different exit angles.
//
// Limitations:
// - Currently only supports one type of gun (however this
//   can be easily changed).
// - Doesn't properly handle multiple scattergun hardpoints.
// - Very specific for use with fighters.
//
// Dispite it's limitations, this patch may still be useful
// for mod-makers, as it contains much information regarding
// the manipulation of bullets as they are being fired (a 
// simple variation of this patch is Standoff's gun convergence
// patch).
//
// Mario Brito
// (HCl)
// mbrito@dei.uc.pt


#include "stdafx.h"
#include "wcsohr.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#ifndef PI
#define PI 3.14159265358979323
#endif

// UE Guns patch
// the struct is incomplete, but for our use it's enough..
typedef struct 
{
void * cpp;
char unknown1[268];
int id;
float damage;
} bullet;

#define TRUE 1
#define FALSE 0

// one of these should be returned on the 
// bullet damage routine (below)
#define DO_DAMAGE TRUE
#define DONT_DO_DAMAGE FALSE

// one of these should be returned on the 
// bullet effect routine (below)
#define HIT_SHIELD_EFFECT FALSE
#define DID_DAMAGE_EFFECT TRUE

// fn1 is called twice for each bullet that hits
// a capship component: first time to determine 
// which visual effect to show, ("hit shield" or 
// "did damage" effects), second time will check if
// it'll do damage to the capship component.
// This flag is used to control this.
// ---
// This is not how WCP usually works, but i wanted
// to avoid hooking 2 different functions. It works tho.
int flag = 0;

int __stdcall guns(bullet * b)
{
// Example:
// Every weapon damages capship components.
// Cloudbursts do double damage to components ONLY!
// Remember that this routine is called after it's been
// determined that b hit a capship component.
/*
if (!flag)
	{
	flag = 1;         // effect routine
	return DID_DAMAGE_EFFECT;
	}
else
	{
	flag = 0;         // damage routine
	if (b->id == 9)
		b->damage *= 2.0f;
	return DO_DAMAGE;
	}
*/

// UE guns patch: usual plasma guns do damage, other guns don't.
// Viper cannon now does half damage to capship components.
if (!flag)
	{
	flag = 1;         // effect routine
	// added Viper and id=15 to "damage" effect 
	// (only 6 and 16 were in the EXE, so 15 does damage but still showed the shield
	// effect on Origin's EXE)
	if (b->id == 2 || b->id == 6 || b->id == 15 || b->id == 16)  
		return DID_DAMAGE_EFFECT;
	else
		return HIT_SHIELD_EFFECT;
	}
else
	{
	flag = 0;			// damage routine
	
	if (b->id == 2)		// UE's Viper cannon
	{
		b->damage /= 2.0f;
		return DO_DAMAGE;
	}
	// plasmas (usually hardcoded, but i intercepted the original code)
	else if (b->id == 6 || b->id == 15 || b->id == 16)
		return DO_DAMAGE;
	else
		return DONT_DO_DAMAGE;
	}

}



// TRE patch 
void tre(void)
{
void (*LoadTre)(char*,int) = (void(*)(char*, int)) 0x4166c0;
int * selectedRes = (int *) 0x43d3c6;

char func[20];
char treName[20];
int number;

func[0] = 0;

// loadtre lines were after the file read

LoadTre("mission.tre", 2);
LoadTre("data.tre", 1);
LoadTre("speech.tre", 1);
LoadTre("ifcomms.tre", 1);
LoadTre("silence.tre", 2);
LoadTre("music.tre", 1);


FILE * f = fopen ("ue_tre.cfg", "rt");
if (f)
	{
	while (1)
		{
		fscanf (f, "%s %s %d\n", func, treName, &number);
		if (!strcmp(func, "end")) break;
		else if (!strcmp(func, "all") || (atoi(func) == *selectedRes))  
			LoadTre (treName, number);
		}
	fclose(f);
	}

}

bool menu = false;




// UE's Scatter Gun patch
typedef struct
{
	float x,y,z;
} VECTOR;

typedef struct
{
	VECTOR v1;
	int x;
	VECTOR v2;
	int y;
	VECTOR v3;
	int z;
} MATRIX34;

typedef struct
{
	VECTOR v1,v2,v3;
} MATRIX33;


VECTOR v[5];
VECTOR * o;
int vector_number = 0;

int scatter_counter = 0;
void * scatter_return = (void *) 0x4403a0;
void * scatter_skip_return = (void *) 0x43fffb;
void * scatter_skip_return2 = (void *) 0x4401ab;
void * scatter_skip_call = (void *)0x40dbd0;
void * scatter_rotate_return = (void *)0x40d51f;

#define SCATTER_INDEX 9  // cloudburst is Scatter Gun now
#define INITIAL_SPREAD 0

#define SCATTER_ANGLE (PI/51.0)

// This function actually makes the bolts rotate ( "scatter" )
void rot(MATRIX34 * src, MATRIX34 * dest, int c)
{
	MATRIX33 m;

	memcpy (&(m.v1), &(src->v1), sizeof(VECTOR));	// copy vectors to temp location
	memcpy (&(m.v2), &(src->v2), sizeof(VECTOR));
	memcpy (&(m.v3), &(src->v3), sizeof(VECTOR));

	dest->x = src->x;  // copy the start position
	dest->y = src->y;
	dest->z = src->z;


	switch (c)
	{
	case 4:
		dest->v1.x = m.v1.x;
		dest->v1.y = m.v1.y * cos(-SCATTER_ANGLE) - m.v1.z * sin(-SCATTER_ANGLE);
		dest->v1.z = m.v1.y * sin(-SCATTER_ANGLE) + m.v1.z * cos(-SCATTER_ANGLE);

		dest->v2.x = m.v2.x;
		dest->v2.y = m.v2.y * cos(-SCATTER_ANGLE) - m.v2.z * sin(-SCATTER_ANGLE);
		dest->v2.z = m.v2.y * sin(-SCATTER_ANGLE) + m.v2.z * cos(-SCATTER_ANGLE);

		dest->v3.x = m.v3.x;
		dest->v3.y = m.v3.y * cos(-SCATTER_ANGLE) - m.v3.z * sin(-SCATTER_ANGLE);
		dest->v3.z = m.v3.y * sin(-SCATTER_ANGLE) + m.v3.z * cos(-SCATTER_ANGLE); 

		break;


	case 3:
		dest->v1.x = m.v1.x;
		dest->v1.y = m.v1.y * cos(SCATTER_ANGLE) - m.v1.z * sin(SCATTER_ANGLE);
		dest->v1.z = m.v1.y * sin(SCATTER_ANGLE) + m.v1.z * cos(SCATTER_ANGLE);

		dest->v2.x = m.v2.x;
		dest->v2.y = m.v2.y * cos(SCATTER_ANGLE) - m.v2.z * sin(SCATTER_ANGLE);
		dest->v2.z = m.v2.y * sin(SCATTER_ANGLE) + m.v2.z * cos(SCATTER_ANGLE);

		dest->v3.x = m.v3.x;
		dest->v3.y = m.v3.y * cos(SCATTER_ANGLE) - m.v3.z * sin(SCATTER_ANGLE);
		dest->v3.z = m.v3.y * sin(SCATTER_ANGLE) + m.v3.z * cos(SCATTER_ANGLE);

		break;


	case 2:
		dest->v1.x = m.v1.x * cos(SCATTER_ANGLE) + m.v1.z * sin(SCATTER_ANGLE);
		dest->v1.y = m.v1.y;
		dest->v1.z = - m.v1.x * sin(SCATTER_ANGLE) + m.v1.z * cos(SCATTER_ANGLE);

		dest->v2.x = m.v2.x * cos(SCATTER_ANGLE) + m.v2.z * sin(SCATTER_ANGLE);
		dest->v2.y = m.v2.y;
		dest->v2.z = - m.v2.x * sin(SCATTER_ANGLE) + m.v2.z * cos(SCATTER_ANGLE);

		dest->v3.x = m.v3.x * cos(SCATTER_ANGLE) + m.v3.z * sin(SCATTER_ANGLE);
		dest->v3.y = m.v3.y;
		dest->v3.z = - m.v3.x * sin(SCATTER_ANGLE) + m.v3.z * cos(SCATTER_ANGLE);

		break;

	case 1:
		dest->v1.x = m.v1.x * cos(-SCATTER_ANGLE) + m.v1.z * sin(-SCATTER_ANGLE);
		dest->v1.y = m.v1.y;
		dest->v1.z = - m.v1.x * sin(-SCATTER_ANGLE) + m.v1.z * cos(-SCATTER_ANGLE);

		dest->v2.x = m.v2.x * cos(-SCATTER_ANGLE) + m.v2.z * sin(-SCATTER_ANGLE);
		dest->v2.y = m.v2.y;
		dest->v2.z = - m.v2.x * sin(-SCATTER_ANGLE) + m.v2.z * cos(-SCATTER_ANGLE);

		dest->v3.x = m.v3.x * cos(-SCATTER_ANGLE) + m.v3.z * sin(-SCATTER_ANGLE);
		dest->v3.y = m.v3.y;
		dest->v3.z = - m.v3.x * sin(-SCATTER_ANGLE) + m.v3.z * cos(-SCATTER_ANGLE);

		break;

	default:
		__asm int 3
	}


}


// This function handles the bolts rotation 
void __declspec (naked) scatter_rotate()
{

	__asm cmp scatter_counter, 0
	__asm jz dont_rotate

	__asm lea esi, dword ptr[eax+0x30]

	__asm pushad
	
	__asm push scatter_counter
	__asm push edi
	__asm push esi
	__asm call rot
	__asm add esp, 12

	__asm popad


	__asm jmp scatter_rotate_return


dont_rotate:
	__asm lea esi, dword ptr[eax+0x30]
	__asm repz movsd
	__asm jmp scatter_rotate_return
}


// This function prevents Scatter bolts from firing when the player is on full guns mode
void __declspec (naked) scatter_skip()
{

	__asm cmp ecx, SCATTER_INDEX
	__asm jz skip_scatter_bolts

	__asm call scatter_skip_call
	__asm jmp scatter_skip_return

skip_scatter_bolts:
	__asm pop eax
	__asm jmp scatter_skip_return2
}


// This function handles the creation of the 5 Scatter bolts when the weapon is on single fire
void __declspec (naked) scatter_gun()  // naked context!
{
	__asm pusha
	
	__asm cmp dword ptr[esi], SCATTER_INDEX	
	__asm jnz leave_scatter

	__asm cmp scatter_counter, 0
	__asm jnz do_scatter_bolt
	
	__asm mov scatter_counter, 5  // reset counter, 5 bolts: center, up, down, left, right
	
	// build vector list (original first) from ... esi+8 [(esi+50) + (ebp-c)]
	__asm mov o, esi
	__asm add o, 4
	
	// LIMIT: Only first hardpoint... :(
	//__asm mov vector_number, dword ptr [esi+0x50]
	//__asm add vector_number, dword ptr [ebp-0xc]
	//o += vector_number;
	__asm mov eax, o
	__asm add o, 4
	__asm mov dword ptr[eax], 1
	
	v[0].x = o->x;
	v[0].y = o->y;
	v[0].z = o->z;
	
	v[1].x = o->x - INITIAL_SPREAD;
	v[1].y = o->y;
	v[1].z = o->z;

	v[2].x = o->x + INITIAL_SPREAD;
	v[2].y = o->y;
	v[2].z = o->z;

	v[3].x = o->x;
	v[3].y = o->y - INITIAL_SPREAD;
	v[3].z = o->z;

	v[4].x = o->x;
	v[4].y = o->y + INITIAL_SPREAD;
	v[4].z = o->z;

	// update bolt counter at ebp-10h
	__asm add dword ptr [ebp-0x10], 4	// 4+1 = 5 bolts

do_scatter_bolt:
	__asm mov dword ptr [esi+0x50],0
	__asm mov dword ptr [ebp-0xc],0

	// Decrement counter..
	__asm dec scatter_counter

	// Update pos vector at "o"
	// This will trash ESI and EDI (compiler optimization) but what the hell...
	// we'll pop all the registers next ;)
	memcpy(o, v+scatter_counter, sizeof(VECTOR));

	__asm cmp scatter_counter, 0
	__asm jnz leave_scatter
 	__asm mov dword ptr [ebp-0xc],5


leave_scatter:
	__asm popa
	__asm mov eax, dword ptr [ebp-0x14]
	__asm test eax, eax
	__asm jmp scatter_return

}







BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
void (*temp)(void) = tre;
char treBridge[] = {0x90, 0x68, 0xa8, 0xac,0x43,0, 0x68, 0,0,0,0,0xc3};
memcpy((void *)0x43ac4e, treBridge, 12);
memcpy((void *)0x43ac55, &temp, 4);

temp = (void(*)(void))guns;
char gunsBridge1[] = {0x53, 0xff, 0x15, 0xd0,0x66,0x4d,0, 0xa, 0xc0, 0x74, 0x10, 0xeb, 0x2};
char gunsBridge2[] = {0x53, 0xff, 0x15, 0xd0,0x66,0x4d,0, 0xa, 0xc0, 0x75, 0x36, 0xeb, 0x3};
memcpy((void *)0x41c109, gunsBridge1, 13);
memcpy((void *)0x41cce0, gunsBridge2, 13);
memcpy((void *)0x4d66d0, &temp, 4); // i think 4d66d0 can be safely overwritten


// install the Scatter Gun patch
// - 5 bolts function
temp = (void(*)(void)) ((unsigned int)scatter_gun - (unsigned int)0x4403a0);
char scatter_bridge[] = {0xe9,0,0,0,0};
memcpy (scatter_bridge+1, &temp, 4);
memcpy ((void*)0x44039b, scatter_bridge, 5);
scatter_counter = 0;
// - skip function (full guns)
temp = (void(*)(void)) ((unsigned int)scatter_skip - (unsigned int)0x43fffb);
memcpy (scatter_bridge+1, &temp, 4);
memcpy ((void*)0x43fff6, scatter_bridge, 5);
//  - bolt rotation function
temp = (void(*)(void)) ((unsigned int)scatter_rotate - (unsigned int)0x40d51f);
memcpy (scatter_bridge+1, &temp, 4);
memcpy ((void*)0x40d51a, scatter_bridge, 5);



HMODULE lib = NULL;
void (*doMenu)(void);
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		
		if (!menu)
			{
			// show menu
			lib = LoadLibrary("wcsohr_menu.dll");
			doMenu = (void (*)()) GetProcAddress(lib,"doMenu");
			if (doMenu)
				doMenu();
			else
				MessageBox(0,"Error loading menu!","ERROR",MB_ICONERROR);
			FreeLibrary(lib);
			menu = true;
			}

		break;
		
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
		break;
    }
    return TRUE;
}


// This is an example of an exported variable
WCSOHR_API int nWcsohr=0;

// This is an example of an exported function.
WCSOHR_API int fnWcsohr(void)
{
	return 42;
}

// This is the constructor of a class that has been exported.
// see wcsohr.h for the class definition
CWcsohr::CWcsohr()
{ 
	return; 
}

